import { execSync } from 'node:child_process';
export const UNIVERSAL_TARGETS = {
    'universal-apple-darwin': ['aarch64-apple-darwin', 'x86_64-apple-darwin'],
};
export const AVAILABLE_TARGETS = [
    'aarch64-apple-darwin',
    'aarch64-linux-android',
    'aarch64-unknown-linux-gnu',
    'aarch64-unknown-linux-musl',
    'aarch64-pc-windows-msvc',
    'x86_64-apple-darwin',
    'x86_64-pc-windows-msvc',
    'x86_64-unknown-linux-gnu',
    'x86_64-unknown-linux-musl',
    'x86_64-unknown-freebsd',
    'i686-pc-windows-msvc',
    'armv7-unknown-linux-gnueabihf',
    'armv7-unknown-linux-musleabihf',
    'armv7-linux-androideabi',
    'universal-apple-darwin',
    'riscv64gc-unknown-linux-gnu',
    'powerpc64le-unknown-linux-gnu',
    's390x-unknown-linux-gnu',
    'wasm32-wasi-preview1-threads',
];
export const DEFAULT_TARGETS = [
    'x86_64-apple-darwin',
    'aarch64-apple-darwin',
    'x86_64-pc-windows-msvc',
    'x86_64-unknown-linux-gnu',
];
export const TARGET_LINKER = {
    'aarch64-unknown-linux-musl': 'aarch64-linux-musl-gcc',
    'riscv64gc-unknown-linux-gnu': 'riscv64-linux-gnu-gcc',
    'powerpc64le-unknown-linux-gnu': 'powerpc64le-linux-gnu-gcc',
    's390x-unknown-linux-gnu': 's390x-linux-gnu-gcc',
};
const CpuToNodeArch = {
    x86_64: 'x64',
    aarch64: 'arm64',
    i686: 'ia32',
    armv7: 'arm',
    riscv64gc: 'riscv64',
    powerpc64le: 'ppc64',
};
export const NodeArchToCpu = {
    x64: 'x86_64',
    arm64: 'aarch64',
    ia32: 'i686',
    arm: 'armv7',
    riscv64: 'riscv64gc',
    ppc64: 'powerpc64le',
};
const SysToNodePlatform = {
    linux: 'linux',
    freebsd: 'freebsd',
    darwin: 'darwin',
    windows: 'win32',
};
export const UniArchsByPlatform = {
    darwin: ['x64', 'arm64'],
};
/**
 * A triple is a specific format for specifying a target architecture.
 * Triples may be referred to as a target triple which is the architecture for the artifact produced, and the host triple which is the architecture that the compiler is running on.
 * The general format of the triple is `<arch><sub>-<vendor>-<sys>-<abi>` where:
 *   - `arch` = The base CPU architecture, for example `x86_64`, `i686`, `arm`, `thumb`, `mips`, etc.
 *   - `sub` = The CPU sub-architecture, for example `arm` has `v7`, `v7s`, `v5te`, etc.
 *   - `vendor` = The vendor, for example `unknown`, `apple`, `pc`, `nvidia`, etc.
 *   - `sys` = The system name, for example `linux`, `windows`, `darwin`, etc. none is typically used for bare-metal without an OS.
 *   - `abi` = The ABI, for example `gnu`, `android`, `eabi`, etc.
 */
export function parseTriple(rawTriple) {
    if (rawTriple === 'wasm32-wasi' ||
        rawTriple === 'wasm32-wasi-preview1-threads' ||
        rawTriple.startsWith('wasm32-wasip')) {
        return {
            triple: rawTriple,
            platformArchABI: 'wasm32-wasi',
            platform: 'wasi',
            arch: 'wasm32',
            abi: 'wasi',
        };
    }
    const triple = rawTriple.endsWith('eabi')
        ? `${rawTriple.slice(0, -4)}-eabi`
        : rawTriple;
    const triples = triple.split('-');
    let cpu;
    let sys;
    let abi = null;
    if (triples.length === 2) {
        // aarch64-fuchsia
        // ^ cpu   ^ sys
        ;
        [cpu, sys] = triples;
    }
    else {
        // aarch64-unknown-linux-musl
        // ^ cpu           ^ sys ^ abi
        // aarch64-apple-darwin
        // ^ cpu         ^ sys  (abi is None)
        ;
        [cpu, , sys, abi = null] = triples;
    }
    const platform = SysToNodePlatform[sys] ?? sys;
    const arch = CpuToNodeArch[cpu] ?? cpu;
    return {
        triple: rawTriple,
        platformArchABI: abi ? `${platform}-${arch}-${abi}` : `${platform}-${arch}`,
        platform,
        arch,
        abi,
    };
}
export function getSystemDefaultTarget() {
    const host = execSync(`rustc -vV`, {
        env: process.env,
    })
        .toString('utf8')
        .split('\n')
        .find((line) => line.startsWith('host: '));
    const triple = host?.slice('host: '.length);
    if (!triple) {
        throw new TypeError(`Can not parse target triple from host`);
    }
    return parseTriple(triple);
}
export function getTargetLinker(target) {
    return TARGET_LINKER[target];
}
export function targetToEnvVar(target) {
    return target.replace(/-/g, '_').toUpperCase();
}
//# sourceMappingURL=target.js.map